/* mod-syscall-xtimefunstruct.c 
 *
 *   Replace the futex 240 system call to return xtime 
 *     as a timeval struct on the stack to user space.
 *   Note that the replacement does no type checking.
 *     because we're not even replacing system calls
 *     with compatable types.
 *
 *   Derived from the no-parameter or result hello-1 example in the
 *   Linux Kernel Module Programmer's Guide for 2.4 Kernel.
 * 
 *   We'll used an existing syscall entry (futex) and replace it
 *
 *   Include file locations
 *      <linux> is /usr/src/linux/include/linux
 *      <asm>   is /usr/src/linux/include/asm
 *      <sys>   is /usr/include/sys
 *
 *   Source file locations
 *      sys_call_table[]    /usr/src/linux/arch/i386/kernel/entry.S
 */

/* kernel module programming */
#ifndef MODULE
#define MODULE
#endif

#ifndef LINUX
#define LINUX
#endif

#ifndef __KERNEL__
#define __KERNEL__
#endif

#define MAX_BUF 80

#include <linux/module.h>    /* needed by all modules */
#include <linux/kernel.h>    /* for KERN_ALERT */
#include <asm/unistd.h>      /* for __NR_futex & other system call numbers */
#include <linux/time.h>      /* for struct timeval */


/* The kernel will fill in the call table at dynamic load (insmod) */
extern void *sys_call_table[];
extern struct timeval xtime;

/* prototypes */
asmlinkage struct timeval my_syscall(void);
int init_module(void);
void cleanup_module(void);

/* My module system call to return xtime on the stack.
 * If the stack value is a reference to memory, then it
 *   won't work.
 */
asmlinkage struct timeval my_syscall(void)
{ printk(KERN_ALERT "my_syscall: entering.\n");
  printk(KERN_ALERT "xtime.tv_sec: %ld  xtime.tv_usec: %ld\n", xtime.tv_sec, xtime.tv_usec);
  printk(KERN_ALERT "my_syscall: leaving.\n");
  return xtime;
}


/* Original system call for NR_futex (240) in sys.c of src/linux/kernel/sys.c
 * The original call is sys_ni_syscall and returns an error number.
 */
asmlinkage long (*original_call_240) (void);

/* Initialize the module at insmod - insert the new system calls */
int init_module(void)
{ printk(KERN_ALERT "syscall-xtimefun module loaded\n");

  /* save the original syscall entry and insert our own entry */
  original_call_240 = sys_call_table[__NR_futex];
  sys_call_table[__NR_futex] = my_syscall;

  /* A non-zero return means init_module failed; module can't be loaded */
  return 0;
}

/* Restore the preious system call at rmmod.
 * Check to see if someone else tried to replace our
 * system calls while we were loaded.
 */
void cleanup_module(void)
{ if (sys_call_table[__NR_futex] != my_syscall )
  { printk(KERN_ALERT "syscall module entry has been corrupted\n");
  }

  /* restore the original syscall entry */
  sys_call_table[__NR_futex] = original_call_240;

  printk(KERN_ALERT "syscall-xtimefun module unloaded\n");
}

MODULE_LICENSE("GPL");
